function [ model, trainerr ] = train( model, urbilder, bilder, sampleclass, trainparams, eps, varargin)
%
%
% function [model, trainerr] = train(model, urbilder, bilder, sampleclass, trainparams, eps, varargin)
%
%
%   Train perceptron class training algorithm based on the Rprop.
%
%   See: C. Igel and M. Hsken, "Improving the Rprop Learning Algoritm",
%        Proceedings of the 2. ICSC Int. Symposium on Neural Computation (NC 2000)
%        pp. 115-121, ICSC Academic Press, 2000
%
  
%% Joerg Wichard 2002

global entooldrawit

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Divide Data in Training and Test

  if isempty(trainparams)
    trainparams = get(model, 'trainparams');
  end
  
  nr_train = length(bilder);
  [ dummy, input_channels ] = size(urbilder);
  
  if ( nr_train ~= dummy )
    error('The number of images and the number of preimages differ');
  end
  
  [indtrain, indtest] = dissemble(sampleclass, nr_train);
  
  %% Scaling the Images and Pre-images
  [urbilder, uscalefacs] = scale_data(urbilder, model.data_scale, model.urbilder_scalefacs);
  [bilder, bscalefac] = scale_data(bilder, model.data_scale, model.bilder_scalefac);
  
  model.urbilder_scalefacs = uscalefacs;
  model.bilder_scalefac = bscalefac;
  
  %% Traindata
  urbilder_train = urbilder(indtrain,:);
  bilder_train = bilder(indtrain);
  
  %% Testdata
  urbilder_test = urbilder(indtest,:);
  bilder_test = bilder(indtest);
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %% Initialize Perceptron 
  
  [ m,n ] = size(  model.tmp_topo );
  if ( m ~= 1 ) 
    error('Inconsistent topologie-parameter for perceptron class');
  end 

  %% Topologie parameter
  model.topo = [ input_channels, model.tmp_topo, 1 ];
  model.layers = ( length(model.topo ) - 1);

  %% Initializing weights: Following Y. LeCun et al. 'Efficient BackProp' in 
  %% Orr and Mller 'Neural Networks, Tricks and Trade', Springer (1998)
  %%
  %% If a neuron has 'm' input-units, the weights are randomly drawn with zero mean
  %% and a stadard deviation of 1/sqrt(m).
  for l=1:model.layers 
    model.w{l}  = (1.0/sqrt((1+model.topo(l))))*randn( (1+model.topo(l)), model.topo(l+1) );
  end	
  
  %% Training the Model
  [ model, terr, epsi ] = train_perceptron_rprop(urbilder_train, bilder_train, trainparams, model);
  
  %% Calculating the test-error
  [ y_train ] = calc_train( model, urbilder_train );
  trainerr = mean( (y_train - bilder_train).^2 );
  
  if ( entooldrawit )
    if( ~isempty(urbilder_test) )
      
      [ y_test  ] = calc_train( model, urbilder_test );
      
      figure(1)
      subplot(1,2,1);
      plot(terr);
      axis square;
      grid on;
      
      subplot(1,2,2)
      plot([-1; 1],[-1; 1],'k');
      hold on;
      plot([(-1-epsi); (1-epsi)],[-1; 1],'k');
      plot([(-1+epsi); (1+epsi)],[-1; 1],'k');
      plot(bilder_train,y_train,'.b');
      plot(bilder_test,y_test,'.r');
      axis([-1 1 -1 1]);
      axis square;
      grid on;
      hold off;
      
      drawnow
      
    else
      figure(1)
      subplot(1,2,1);
      plot(terr);
      axis square;
      grid on;
      
      subplot(1,2,2)
      plot([-1; 1],[-1; 1],'k');
      hold on;
      plot([(-1-epsi); (1-epsi)],[-1; 1],'k');
      plot([(-1+epsi); (1+epsi)],[-1; 1],'k');
      plot(bilder_train,y_train,'.b');
      axis([-1 1 -1 1]);
      axis square;
      grid on;
      hold off;
      
      drawnow
    end
    
  end
  %% 
  
  
function [ model, terr, epsi ] = train_perceptron_rprop( urbilder, bilder, trainparams, model)
%
%
%  1. Order Gradient decend,  Rprob 
%
  
%% Calculating the scales error loss margin
  data_spread = ( max(max(urbilder)) - min(min(urbilder)) );
  epsi = (trainparams.error_loss_margin*data_spread);
  
  [ data_length, dummy ] = size(urbilder);
  
  %% Die Datenstructur zum Rprop bereitstellen
  ts = create_train_struct(model, trainparams);
  
  %% Trainings-daten
  xin   = urbilder;
  xreal = bilder;
  
  [ m, dummy ] = size( xin );
  
  %% Die verschiedenen Fehler
  terr = zeros(1,trainparams.rounds);
  
  err_new = 0;
  for iter=1:trainparams.rounds
    
    x =  xin;
    ts.xtmp{1} = [ x, model.offsett*ones(m,1) ];
    ts.y{1} = x;
    ts.dev{1} = ones(m,model.topo(1));

    wdecay = 0.0;
    for l=1:model.layers
      x = tanh(ts.xtmp{l}*model.w{l}) + (model.linear_factor*(ts.xtmp{l}*model.w{l}));
      ts.xtmp{l+1} = [ x, model.offsett*ones(m,1)];
      ts.y{l+1} = x;
      wdecay = ( wdecay + sum(sum( (model.w{l}).^2)) );
      ts.dev{l+1} = ( 1.0 - x.^2) + model.linear_factor*ones(size(x));
    end	
    
    %% Den Epsilon-Insensitiven-Fehler
    ecount = ( (abs( xreal - x )) > epsi );
    
    %% Der Fehler am Netzausgang 
    ts.delta{(model.layers + 1)} = trainparams.decay*wdecay + (ecount).*( -((xreal - x) - epsi ));
    
    err_old = err_new;
    err_new = (sum(ts.delta{(model.layers + 1)}.^2));
    
    %% Den Fehler zurueckpropagieren
    for l=model.layers:-1:1
      ts.delta{l} = (ts.delta{(1+l)}*transpose(model.w{l}(1:(end-1),:))).*ts.dev{l};
    end
    
    %% Den einzelnen Gradienten berechnen
    for l=1:model.layers
      ts.dnew{l} = ( (transpose([ ts.y{l}, model.offsett*ones(m,1) ])*ts.delta{l+1} ) + ...
		     trainparams.decay*model.w{l});
    end 
    
    %% Den Gradienten berechnen
    for l=1:model.layers
      
      %% Die Richtungsinformation 
      dtmp = sign( ts.dold{l} .* ts.dnew{l} );
      
      %% Die alte Schrittweite, zum zurcknehmen
      ts.dw_old{l} = ts.dw{l}; 
      
      %% Die Schrittweite adaptieren
      ts.dw{l} = (  trainparams.mrate_grow*( ts.dw_old{l}.*(dtmp == 1)) ...
		    + trainparams.mrate_shrink*(ts.dw_old{l}.*(dtmp == -1)) ...
		    + ts.dw_old{l}.*(dtmp == 0) );
           
      %% Die Schrittweite festlegen, abhngig vom Fehler
      if ( err_new > err_old )
	
	ts.dstep{l} = (    sign(ts.dnew{l}).*( ts.dw{l}.*(dtmp == 1)) ...
			   + sign(ts.dnew{l}).*( ts.dw{l}.*(dtmp == 0)) ...
			   - ts.dw_old{l}.*( dtmp == -1) );
      else
	ts.dstep{l} = zeros( (1+model.topo(l)), model.topo(l+1) );
	
	ts.dstep{l} = (    sign(ts.dnew{l}).*( ts.dw{l}.*(dtmp == 1)) ...
			   + sign(ts.dnew{l}).*( ts.dw{l}.*(dtmp == 0)) );
      end
      
      model.w{l} = (model.w{l} - ts.dstep{l});
    
      %% Alte Gradienten abspeichern und beim Richtungswechsel die Ableitung Null setzen
      ts.dold{l} = ts.dnew{l};
      itmp = find( dtmp == -1 );
      ts.dold{l}(itmp) = 0;
          
      %% Maximalgewicht
      itmp = find( model.w{l} > trainparams.max_weight );
      model.w{l}(itmp) = trainparams.max_weight;
      ts.dold{l}(itmp) = 0;

      itmp = find( model.w{l} < -trainparams.max_weight );
      model.w{l}(itmp) = -trainparams.max_weight;
      ts.dold{l}(itmp) = 0;
    end
    
    terr(1,iter) =  log(err_new);
    
  end
  
  trained_model = model;
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	
%% Die Datenstructur zum Rprop bereitstellen	

function [ ts ] = create_train_struct(model, trainparams )

  for l=1:model.layers 
    
    %% Die Schrittweite des Gradientenabstiegs
    ts.dw{l}     = (trainparams.mrate_init)*ones( (1+model.topo(l)), model.topo(l+1) );
    ts.dw_old{l} = (trainparams.mrate_init)*ones( (1+model.topo(l)), model.topo(l+1) );
    ts.dstep{l}  = (trainparams.mrate_init)*ones( (1+model.topo(l)), model.topo(l+1) );
    
    %% Die Gradienten 
    ts.dold{l} = zeros( (1+model.topo(l)), model.topo(l+1) );
    ts.dnew{l} = ones( (1+model.topo(l)), model.topo(l+1) );
    
    %% Zum Zurckrechnen der Gradienteninformtion
    ts.delta{l+1} = zeros( 1, model.topo(l+1) );
    
    %% Die Zwischenergebnisse
    ts.y{l+1} = zeros( 1, model.topo(l+1) );
    
    %% Die Zwischenergebnisse mit On-Neuron
    ts.xtmp{l+1} = zeros( 1, ( 1 + model.topo(l+1) ));
    
    %% Die Ableitungen
    ts.dev{l+1} = zeros( 1, model.topo(l+1) );
    
  end
  
  %% Die Zwischenergebnisse des Ausgabeneurons
  ts.xtmp{1} = zeros( 1, ( 1 + model.topo(1) ));

  ts.y{1}   = ones(1,model.topo(1));

  ts.dev{1} = ones(1,model.topo(1));

  ts.delta{1} = zeros(1,model.topo(1));

   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Calculating the epsilon insensitiv error

function [ err ] = calc_error( xreal, xout, epsi )

decide = abs( xreal - xout ) > epsi;

err = mean(( decide.*( abs(xreal - xout) - epsi )).^2);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Calc Model during the training, without scaling

function [ xout ] = calc_train( model, xin )

[ m, dummy ] = size( xin );

for l=1:model.layers
  xin = [ xin, model.offsett*ones(m,1) ];
  xin = (tanh(xin*model.w{l}) + model.linear_factor*(xin*model.w{l}));
end	

xout = xin;
