function [rr_net,iter,cinit,nhid] = prbfn_train(p, t, k, opts)
% rr_net = prbfn_train(p, t)
% Description: This is the train function for ridge and
% rbfs. The algorithm works as follows:
% 1. Split the input space into regions in a greedy fashion to reduce 
%    an objective function.
% 2. It finds the best centers from data, i.e. has most
%    of the patterns, and use them as starting point for
%    centers and radiuses.
% 3. Use gradient descent for finding the local minimum
% 4. Possibliy a pruning using the evidence of LRT for model selection.
% The regularization parameters (see below) are for weight decay. When prunning is
% performed (the default in regression problems) usually regularization parametrs may set to zero.
% Inputs:
% p - a matrix of patterns, each column is a pattern.
% t - a matrix of targets, each column is one target.
% k - the maximum number of ridges and rbfs, if k is negative the algorithm determines k from the data.
% opts.reg1 - first layer regularization param in the range[1-e5,0.5]
% opts.reg2 - output layer regularization param range as above, output of percpetrons.
% opts.reg3 - output layer regularization param as above, output of RBFs.
% opts.regr - Rbf width regularization param, this makes sure the width is not small. If regr>=1 then the radius
%             becomes the maximum distance between clusters centers. It is optimized later in the optimization process.
% opts.err_tar - requested error target (MSE), if err_atr is negative the error target is set from the data automatically.
% opts.alg_tye - optimization type use 1 for conjugate gradient.
% opts.out_type - unit output type 1-logsig,2-linear. Default logsig.
% opts.hid_type - selection of the projection units from logsig or tanh: 1 logsig,2-tanh. Default logsig.
% opts.prune_val - the amount to prune. This is valid only for regression when the dimension > 1. When prune_val in[0,1] this is the amount
% to prune from all the parameters. When it is larger then one this number denotes the maximum number of parameters to
% prune.
% opts.model_sel_type: This must be set inconjunction with the previous value. The type of the model selection out of the range 
% of model that are specifed in prune_val are:BIC = 2, SSE/(n-p) =1, LRT(0.95), LRT(0.975),LRT(0.99).
% Outputs:
% rr_net - the net which include matrices for weights, 
%          centers and radiuses. This data structure
% Author: Shimon Cohen
% Copyrights(C) Shimon Cohen
opts.full_evid = 0;
% mcc -x -h <fname> to compile
[rr_net,cinit, max_iter,err_tar,opts] = prbfn_init(p,t,k, opts);
if (isempty(opts) | opts.err_tar < 0)
   [rr_net,iter] = prbfn_vconj_grad(rr_net, p, t, max_iter, err_tar, opts);    
else
   [rr_net,iter] = prbfn_conj_grad(rr_net, p, t, max_iter, err_tar, opts);
end
nhid = rr_net.nridges + rr_net.nrbfs;


function [rr_net,cinit,max_iter, err_tar,opts,ensemble] = prbfn_init(p,t,k,opts)
% [ rr_net, cinit,max_iter,err_tar,ensemble] = prbfn_init(p,t,k,opts)
% initializes a net form the patterns p, targets t, 
% k-hiddens and opts the options
cinit=[];
ensemble = [];
[dim, npats] = size(p);
[C, npats] = size(t);
cluster_alg = 1;
init_flag = 0;
regFlag = 0;
if (C == 1 & (length(find(t==1))+length(find(t==0)) ~=npats))
   regFlag = 1;
end
if (regFlag)
    rr_net.output_type = 2;
else
    rr_net.output_type = 1;
end
rr_net.hid_type = 1;
rr_net.err_type = 2;
err_tar = 0.05;
max_iter = 0;
sort_flag = 0;
verbose = 0;
if (~isempty(opts))
   if (isfield(opts, 'verbose'))
      verbose = opts.verbose;
   end
    if (isfield(opts,'output_type') & ~isempty(opts.output_type))
      rr_net.output_type = opts.output_type;
   end
   if (isfield(opts,'hid_type') & ~isempty(opts.hid_type))
      rr_net.hid_type = opts.hid_type;
   end
   if (~isempty(opts.err_tar) & opts.err_tar >= 0)
      err_tar = opts.err_tar;
   end
   if (~isempty(opts.max_iters))
      max_iter = opts.max_iters;   
   end
   if (~isfield(opts, 'reg1') | isempty(opts.reg1))
      opts.reg1 =0;
      if (verbose)
        disp('Warnning prbfn_init: setting to 0 reg1');
      end
   end
   if (~isfield(opts, 'reg2') | isempty(opts.reg2))
      opts.reg2 = 0;
      if (verbose)
        disp('Warnning prbfn_init: setting to 0 reg2');
      end
   end
   if (isfield(opts, 'err_type') & ~isempty(opts.err_type))
      rr_net.err_type = opts.err_type;     
   end
   if (~isfield(opts, 'reg3') | isempty(opts.reg3))
      opts.reg3 = 0;
      if (verbose)
        disp('Warnning prbfn_init: setting to 0 reg3');
      end
   end
   if (~isfield(opts, 'regr') | isempty(opts.regr))
      opts.regr = 0;
      if (verbose)
        disp('Warnning prbfn_init: setting to 0 regr');
      end
   end
   if (~isfield(opts, 'alg_type') | isempty(opts.alg_type))
        opts.alg_type = 1;
   end
else
   opts.reg1 = 0;
   opts.reg2 = 0;
   opts.reg3 = 0;
   opts.regr = 0;
end
opts.verbose = verbose;
if (isempty(opts) | ~isfield(opts,'cluster_alg') | ...
      isempty(opts.cluster_alg) | ~isfield(opts,'per_sel_type') | ...
      isempty(opts.per_sel_type)| ~isfield(opts,'clust_dist') |...
      isempty(opts.clust_dist))
   opts.cluster_alg = 13;
   rr_net.vec_rad = 0;
   [m,r,cm,ci,clusts] = prbfn_clust(rr_net,p,t,k,opts.err_tar,opts);
   flgs = [];
   if (regFlag==0)
      r = vecdist(m);
      opts.per_sel_type = 6;
      [M2,rbf_inds, mlp_inds] = get_rbfs_from_likelihood_ratio(p,m,r);   
      rr_net = create_net(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts);
   else
      opts.per_sel_type = 7;
      if (opts.regr >= 1)
         r = maxdist(m) * opts.regr;
         r = ones(dim,1) * r;
         opts.regr = 0;
      end
      rr_net = create_net_from_clusts(rr_net,clusts,p,t,m,r,opts);
      if (opts.regr < 0)
         [y,z,a1,a2] = prbfn_out(p, rr_net,0);
         x = z';
         x = x(2:end,:);
         [w,wo,reg] = blr(x,t);
         rr_net.w2(1) = wo;
         rr_net.w2(2:end) = w';
         opts.regr = 0;
      end
   end
   
   if (rr_net.vec_rad)
      [rr_net,ensemble] = prbfn_prune(p,t, rr_net, opts);
   else
      [rr_net,lr] = compute_rads(rr_net,p,t,p,t); 
   end
   %   rr_net = compute_cv_rads(rr_net,p,t);
   return;
end

if (~isempty(opts))
   cluster_alg = opts.cluster_alg;
end

ntrys = 1;
rr_net.vec_rad = 0;
if (cluster_alg == 1)
   [m,r,cm] = scmean(p, t, k, opts.max_iters1);
elseif (cluster_alg == 2)
   [m,r,cm] = skmeans(p,k, ntrys,t);
elseif (cluster_alg == 3)
   [m,r, cm, dist, ci] = tmean(p,t, k);
elseif (cluster_alg == 4)
   [m,r,cm,cinit] = tclust(p,t,k,0,opts.err_tar);
elseif (cluster_alg == 5)
   [m, r, cm,c] = tree_clust(p, t, k,5);
elseif (cluster_alg == 6)
   %conf.minmem = [6];
   %conf.scales = [4 6 8 10 12];
   %[m,r,w, info,conf] = rbf_rt_1(p, t', conf);
   root = tree_max_split(p,t, 5,0);
   root = tree_prune(root, k,opts.crit_type);
   [m,r,cm] = tree_get_rbfs(root);
   [dimR,nr] = size(r);
   rr_net.vec_rad = (dimR>1);
elseif (cluster_alg == 7)
   reg_tree_err = opts.err_tar;
   if (isfield(opts, 'regr'))
      reg_tree_err = opts.regr;
   end

   [m,r,cm] = cart_train(p,t,k);
   [dimt,k] = size(m);
   %   r = vecdist(m);
   rr_net.vec_rad = 1;
elseif (cluster_alg == 8)
   [m,c,r] = ml_clust(p,k,1);
   cm = ones(1,k);
elseif (cluster_alg == 9)
   [m,r,cm,dist] = fkmeans(p,k);   
elseif (cluster_alg == 10)
   %   [m, r, cm,c] = tree_clust(p, t, k,5);
   %   r = vecdist(m);
   [m,r,cm,jval] = cmean(p,t,k);
elseif (cluster_alg == 11)
   [m,r,cm] = tclust(p,t,k,1, opts.err_tar);
elseif (cluster_alg == 12)
   %   rr_net = prbfn_fwd_train(p,t,opts.err_tar,k,opts);
   [clusts,m,r] = prbfn_reg_clusts(p,t,k);
   rr_net = create_net_from_clusts(rr_net,clusts,p,t,m,r,opts);
   rr_net.vec_rad = 1;
   rr_net = prbfn_prune(p,t, rr_net, opts);
   err_tar = opts.err_tar;
   max_iter = opts.max_iters;
   m = rr_net.m;
   r = rr_net.rads;
   return;
elseif (cluster_alg == 13)
   [m,r,cm] = prbfn_clust(p,t,k, opts.err_tar);
   if (regFlag == 1)
      r = max(r);
   else
      r = vecdist(m);
   end
elseif (cluster_alg == 14)
   % RBF only
   clust_dist = opts.clust_dist;
   [m,r,cm] = skmeans(p,k, ntrys,t);
   [dim,K] = size(m);
   flgs = ones(dim,K);
   r = ones(dim,1) * r;
   if (clust_dist>0)
      [M2,cm,rbf_inds, mlp_inds] = prbfn_fit(p,t,m,r);
   else
       mlp_inds = []; rbf_inds=[1:K];
   end
   rr_net = prbfn_create_net(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts,flgs);
   rr_net = prbfn_prune(p,t, rr_net, opts);
   return;
elseif (cluster_alg == 15)
   [m,r,cm] = skmeans(p,k, ntrys,t);
   [dim,K] = size(m);
   flgs = ones(dim,K);
   r = ones(dim,1) * r;
   mlp_inds = [1:K]; rbf_inds=[];
   rr_net = prbfn_create_net(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts,flgs);
   rr_net = prbfn_prune(p,t, rr_net, opts);
   return;  
elseif (cluster_alg == 16)   
   %   [m,r,cm] = prbfn_func_clust_approx(p,t,k);
   [m,r,cm] = clust_func_approx(p,t,k);
else
   error('prbfn unknown clustering algorithm');
end
[dimr,k] = size(r);
cm_num = cm;
M1 = round(k / 2);
M2 = M1;
if (~isfield(opts,'clust_dist') | isempty(opts.clust_dist))
   opts.clust_dist = 2;
end
wgt1 = [];
if (dimr > 1)
   rr_net.vec_rad = 1;
end
if (regFlag & dimr == 1)
   if (opts.regr >= 0)
      r = maxdist(m);
      if (opts.regr > 1)
         r = r * opts.regr;
         opts.regr = 0;
      end
   else
      r = vecdist(m);
      if (opts.regr < -1)
         r = r*(-opts.regr);
      end
      opts.regr = 0;
   end
end

if (opts.per_sel_type == 1)
   [M2,cm,rbf_inds, mlp_inds] = get_rbf_max_spec_num(p,m, opts.clust_dist);
elseif (opts.per_sel_type == 2)
   [M2,cm,rbf_inds, mlp_inds] = get_rbf_num_var_spec(p,m, opts.clust_dist);
elseif (opts.per_sel_type == 3)
   % start with full net and prune half of it
   rr_net = create_net(rr_net,p,t,m,r,[1:k],[1:k],opts);
   init_flag = 1;
   rr_net.wgts_drv = 0;
   rr_net = prbfn_prune(p,t, rr_net, opts);
   return;
elseif (opts.per_sel_type == 4)
   % use normalization test
   [M2,cm,rbf_inds, mlp_inds] = get_nrbf_from_norm_test(p,m,opts.clust_dist);   
elseif (opts.per_sel_type == 5)
     % start with full net and prune half of it
   rbf_net = create_net(rbf_net,p,t,m,r,[],[1:k],opts);

   proj_net = create_net(proj_net,p,t,m,r, [1:k],[],opts);
   %   rr_net.w2(:,k+2:end) = w2(:,2:end);
   init_flag = 1;
   rr_net.wgts_drv = 0;
   [rr_net] = prbfn_merge(p,t, rbf_net,proj_net, opts);
   M2 = rr_net.nrbfs; 
elseif (opts.per_sel_type == 6)
   [M2,rbf_inds, mlp_inds] = get_rbfs_from_likelihood_ratio(p,m,r);   
elseif (opts.per_sel_type == 7)
   [M2,cm,rbf_inds, mlp_inds] = prbfn_fit(p,t,m,r);
end
M1 = k - M2;
if (~isempty(opts))
   if (~isempty(opts.rbf_prior))
      M1 = opts.rbf_prior;
      M2 = k - M1;
      sort_flag = 1;
   end
end

if (init_flag == 0)
   M1 = k - M2;
   if (M1 + M2 < k)
      M1 = M1 + 1;
   elseif(M1 + M2 > k)
      M1 = M1 - 1;
   end
   if (sort_flag == 1)
      [cms, I] = sort(cm);
      rbf_inds = I(end - M2+1:end);
      mlp_inds = I(1:M1);
   end
   rr_net = create_net(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts);
end
rr_net.wgts_drv = 0;
if (regFlag & dimr > 1)
%   rr_net = compute_rads(rr_net, p,t,p,t);
end

if (dimr > 1)
   rr_net.vec_rad = 1;
end

%--------------------------------------------------
function rr_net = create_net_from_clusts(rr_net,clusts,p,t,m,r,opts)
if (isfield(clusts{1}, 'fit'))
   rbf_inds=[];mlp_inds=[];
   for i=1:length(clusts)
      if (strcmp(clusts{i}.fit,'mlp'))
         mlp_inds = [mlp_inds i];
      else
         rbf_inds = [rbf_inds i];
      end
      %            flgs = [flgs clusts{i}.flgs];
   end
else
   [M2,cm,rbf_inds, mlp_inds] = prbfn_fit(p,t,m,r,clusts);
end
rr_net = prbfn_create_net(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts);
if (opts.verbose)
   fprintf(1, 'nridges=%d, nrbfs=%d\n', rr_net.nridges, rr_net.nrbfs);
end

%----------------------------------------------------
function net = create_net(net,p,t,m,r, mlp_inds,rbf_inds,opts)
% creates the actual net
net=prbfn_create_net(net,p,t,m,r, mlp_inds,rbf_inds,opts);
if (opts.verbose)
   fprintf(1, 'nridges=%d, nrbfs=%d\n', net.nridges, net.nrbfs);
end

%----------------------------------------------------
function rr_net = compute_cv_rads(rr_net,p,t)
[cvs, num_folds] = get_cross_valid(p,t, 10);
lrvals=[];
[numC,N] = size(t);
for i=1:num_folds
   trn_idxs = cvs.folds_train_idx{i};
   tst_idxs = cvs.folds_test_idx{i};
   
   p1 = p(:,trn_idxs); t1 = t(:,trn_idxs);
   p2 = p(:,tst_idxs); t2 = t(:,tst_idxs);
   [tnet,lr] = compute_rads(rr_net,p1,t1,p2,t2);
   lrvals = [lrvals lr];
end
% Builds an histogram
minLR = min(lrvals);
numBins = 3;
binSize = (max(lrvals) - minLR)/numBins;
lrsets{numBins+1} = [];
for i=1:length(lrvals)
   id = round((lrvals(i) - minLR)/binSize) + 1;
   lrsets{id} = [lrsets{id} lrvals(i)];
end
set_size = 0;
id = 0;
for i=1:length(lrsets)
   cur_len = length(lrsets{i});
   if (cur_len > set_size)
      id = i;
      set_size = cur_len;
   end
end
lr = mean(lrsets{id});
rr_net.rads = rr_net.rads .* lr;

%----------------------------------------------------
function [net,lr,err] = compute_rads(rr_net, p,t,pt,tt)
% net = compute_rads(rr_net)
% computes best radial for the net.
net = rr_net;
net.w2 = compute_w2(p,t,net);
[y,z,a1,a2] = prbfn_out(pt, net,1);
lrvals = [1:3];
errvals(1) = sumsqrerr(tt,y, net.err_type);
for i=2:3
   net.rads = rr_net.rads * i;
   net.w2 = compute_w2(p,t,net);
   [y,z,a1,a2] = prbfn_out(pt, net,1);
   errvals(i) = sumsqrerr(tt,y, net.err_type);
end

cnt = 0;
while (errvals(2) > errvals(1) | errvals(2) > errvals(3))
   if (errvals(1) > errvals(3) | errvals(2) > errvals(3))
      lrvals(1:2) = lrvals(2:3);
      errvals(1:2) = errvals(2:3);
      i = 3;
      lrvals(3) = 2 * lrvals(3);
   else
      lrvals(2:3) = lrvals(1:2);
      errvals(2:3) = errvals(1:2);
      i = 1;
      lrvals(1) = lrvals(1) /2;
   end
   net.rads = rr_net.rads .* lrvals(i);
   net.w2 = compute_w2(p,t,net);
   [y,z,a1,a2] = prbfn_out(pt, net,1);
   errvals(i) = sumsqrerr(tt,y, net.err_type);  
   cnt = cnt + 1;
   if (cnt > 10)
      break;
   end
end
[a,b,c] = parab_fit(lrvals, errvals);
if (a > 0)
   lrc = - b/(2*a);
else
   lrc = -1;
end
if (lrc > 0)
   net.rads = rr_net.rads .* lrc;
   net.w2 = compute_w2(p,t,net);
   [y,z,a1,a2] = prbfn_out(pt, net,1);
   lrvals(4) = lrc;
   errvals(4) = sumsqrerr(tt,y, net.err_type);  
end
[err,id] = min(errvals);
lr = lrvals(id);
net.rads = rr_net.rads .* lr;
net.w2 = compute_w2(p,t,net);
[y,z,a1,a2] = prbfn_out(pt, net,1);
err = sumsqrerr(tt,y, net.err_type);  


%----------------------------------------------------
function w2 = compute_w2(p,t,net)
[y,z,a1,a2] = prbfn_out(p, net,1);
%if (isfield(opts,'reg2'))
%   z = z + opts.reg2;
%end

w2t = pinv(z) * t';
w2 = w2t';
%-----------------------------------------------------
function [lhrbf, lhpro] = compute_likelihood(p,m,clust,r)
% [lhrbf, lhpro] = compute_likelihood(p,m,clust,r)
% Description: Computes likelihoods of rbf and projection
% unit for a given data set.
%
% Author: Shimon Cohen
[dim, sizeC] = size(clust);
[dim,N] = size(p);
w = m./sqrt(m'*m);

lhrbf = 0;
lhpro = 0;
%---------------------------------------------------------
lhrbf = 1; lhpro =1;
r = sqrt(mean(max(cov(clust))));
%----------------------------------------------------------
r2 = r * r * 2;
sum_pro = 0;
for jj=1:sizeC
   diff = clust(:,jj) - m;
   diff = (diff' * diff) ./ r2;
   vg = exp(-diff);
   vp = w'*clust(:,jj);
   vp = 1 ./ (1 + exp(-vp));
   sum_pro = sum_pro + vp;
   lhpro = lhpro * vp;
   lhrbf = lhrbf * vg;
end
% Compute normalization factors
diff = p - m * ones(1,N);
diff = exp(-sum(diff.*diff,1)./r2);
sum_rbf = sum(diff);
sum_pro = p'*w;
sum_pro = 1./(1+exp(-sum_pro));
sum_pro = sum(sum_pro);
sum_pro = sum_pro^sizeC;
sum_rbf = sum_rbf^sizeC;
if (sum_pro == 0)
    sum_pro = eps;
end
if (sum_rbf==0)
    sum_rbf=eps;
end
lhpro = lhpro/sum_pro;
lhrbf = lhrbf/sum_rbf;

%-------------------------------------------------------------------
function [M2,rbf_idxs, mlp_idxs] = get_rbfs_from_likelihood_ratio(p,m,r)
%[rbf_idxs, mlp_idxs] = get_rbfs_from_likelihood_ratio(p,m,r)
% Description: For each cluster computes likelihood for projection
% and gaussinas units
[clust_mem, clust_idx] = get_clust_mem(p,m);
[dim,K] = size(m);
for ii=1:K
   [lg,lp] = compute_likelihood(p,m(:,ii),clust_mem{ii},r(:,ii));
   if (lp > lg)
      rbf_idxs(ii) = 0;
      mlp_idxs(ii) = 1;
   else
      rbf_idxs(ii) = 1;
      mlp_idxs(ii) = 0;
   end
end
rbf_idxs = find(rbf_idxs > 0);
mlp_idxs = find(mlp_idxs > 0);
M2 = length(rbf_idxs);


%----------------------------------------------------
% Auxilary functions
%-------------------------------------------------------------------
function [nrbf,cm,rbf_inds, mlp_inds] = get_rbf_max_spec_num(p,cnts, spec_thr)
% [nrbf, cm, rbf_inds, mlp_inds] = mlp_rbf_get_rbf_num(m,c,cm)
%
% Author: Shimon Cohen

[dim,k] = size(cnts);
[dim, npats] = size(p);
dist = zeros(1,k);
ones_vec = ones(1,k);
dist_mat = zeros(dim,k);
cm = zeros(1,k);
if (isempty(spec_thr))
   spec_thr = 0.1;
end

% for each pattern finds closest center and adds the distance to it
for i=1:npats
   dmat = p(:,i) * ones_vec;
   dmat = dmat - cnts;
   dmat = dmat .* dmat;
   dmat = sum(dmat);
   [mval, idx] = min(dmat);
   dmat = abs(p(:,i) - cnts(:,idx));
   
   dist_mat(:,idx) = max(dmat,dist_mat(:,idx));
   dist(idx) = dist(idx) + mval;
end

% compute spectrum
for i=1:k
   max_val = max(dist_mat(:,i));
   min_val = min(dist_mat(:,i));
   if (max_val > 0)
      cm(i) = min_val / max_val;
   else
      cm(i) = 1;
   end
end

mlp_inds = find(cm < spec_thr);
rbf_inds = find(cm >= spec_thr);
nrbf = length(rbf_inds);

%-------------------------------------------------------------------------
function [nrbf,cm,rbf_inds, mlp_inds] = get_rbf_num_var_spec(p,cnts, spec_thr)
% [nrbf, cm, rbf_inds, mlp_inds] = get_rbf_num_var_spec(m,c,cm)
% This fuction computes average spectrum rather then maximum for each
% coordinate.
% Author: Shimon Cohen

[dim,k] = size(cnts);
[dim, npats] = size(p);

ones_vec = ones(1,k);
dist_mat = zeros(dim,k);
cm = zeros(1,k);
if (isempty(spec_thr))
   spec_thr = 0.1;
end

% Sets empty matrices
clust_mems{k} = [];

% for each pattern finds closest center and adds the distance to it
for i=1:npats
   dmat = p(:,i) * ones_vec;
   dmat = dmat - cnts;
   dmat = dmat .* dmat;
   dmat = sum(dmat);
   [mval, idx] = min(dmat);
   dmat = p(:,i) - cnts(:,idx);
   dmat = dmat .* dmat;
   cm(idx) = cm(idx) + 1;
   dist_mat(:,idx) = dmat + dist_mat(:,idx);
   clust_mems{idx} = [clust_mems{idx} p(:,i)];
end

% compute spectrum
for i=1:k
   if (cm(i) > 0)
      max_val = max(dist_mat(:,i) ./ cm(i));
      min_val = min(dist_mat(:,i) ./ cm(i)); 
      cm(i) = min_val / max_val;
   else
      cm(i) = 0;
   end
end

mlp_inds = find(cm < spec_thr);
rbf_inds = find(cm >= spec_thr);
nrbf = length(rbf_inds);


%---------------------------------------------------------
function [M2,cm,rbf_inds, mlp_inds] = get_nrbf_from_norm_test(p,m,clust_dist)
% [M2,cm,rbf_inds, mlp_inds] = get_nrbf_from_norm_test(p,m,opts.clust_dist)
% Description: gets the number of rbfs from normality test
[clust_mems, idxs] = get_clust_mem(p,m);
[dim,K] = size(m);
rbf_inds = [];
mlp_inds = [];
for ii=1:K
   a = norm_test(clust_mems{ii});
   cm(ii) = a;
   if (a > clust_dist)
      rbf_inds = [rbf_inds ii];
   else
      mlp_inds = [mlp_inds ii];
   end
end
M2 = length(rbf_inds);

%--------------------------------------------------------------------
function maxrad = maxdist(cnts)
% rads = maxdist(cnts)
% Description: Given a matrix of vectors return distance
% of centers closest cnter
% Inputs:
% cnts - a matrix of vectors, each colum is a vector
% Output:
% rads - vector of distances
% 
% Author: Shimon Cohen
[dim,k] = size(cnts);
ones_vec = ones(1,k);
for c=1:k
   rads(c) = 0;
   cr = cnts(:,c) * ones_vec - cnts;
   cr = sum(cr.*cr,1);
   rads(c) = sqrt(max(cr));
end
if (dim > 0 & k > 0)
   maxrad = max(rads) * ones(1,k);
else
   maxrad = 1;
end

%----------------------------------------------------------
function rads = vecdist(cnts)
% rads = vecdist(cnts)
% Description: Given a matrix of vectors return distance
% of centers closest cnter
% Inputs:
% cnts - a matrix of vectors, each colum is a vector
% Output:
% rads - vector of distances
% 
% Author: Shimon Cohen
[dim,k] = size(cnts);
for c=1:k
   rads(c) = 0;
   for ci=1:k
      if (ci ~= c)
         d = cnts(:,c) - cnts(:,ci);
         d = d' * d;
         if (d < rads(c) | rads(c) == 0)
            rads(c) = d;
         end
      end
   end
end
r=sqrt(rads);

function [cnts, rads, cm,cinit,clusts] = prbfn_clust(net,p, t, k,err_tar,opts)
% [cnts, rads, cm,cinit] = prbfn_clust(p, t, k,err_tar,opts)
% Description: clusterization using mnimimazation on function values
%
% Author: Shimon Cohen
clusts = [];
[dim, N] = size(p);
[tdim, N] = size(t);
bic_flag = 0;
if (k<0)
    k = round(sqrt(N));
    bic_flag = 1;
end
n = 1; 
conf_level_flag = 1;
conf_level = 3.84;
rbf_prior = 1;
full_evid = 1;
if (nargin > 4)
   if (isfield(opts, 'model_sel_type'))
      model_sel = opts.model_sel_type;
      if (model_sel == 4)
         conf_level = 3.84;
      elseif (model_sel == 5)
         conf_level = 5.02;
      elseif (model_sel == 6)
         conf_level = 6.63;
      end
   end
   if (isfield(opts, 'full_evid'))
       full_evid = opts.full_evid;
   end
   if (isfield(opts,'rbf_prior'))
      rbf_prior = opts.rbf_prior;
   end
else
   fprintf(1,'func_clust: using default confidence level of 0.05\n');
   conf_level = 3.84;
end

regFlag = 0;
if (tdim == 1)
   id1 = find(t==1);
   id0 = find(t==0);
   if ((length(id1) + length(id0)) ~= N)
      regFlag = 1;
   end
end
nmin = 5;
cluster.p = p;
cluster.t = t;
yavg = mean(t,2);
diff = t - yavg * ones(1,N);
err1 = sum(diff.*diff,1);
[err,cluster.max_err_id] = max(err1);
cluster.err = sum(err1);
cluster.sse = cluster.err;
cluster.csz = N;
clusts{1} = cluster;
sse = err_tar + 1;
nmin2 = nmin * 2;
chi_val = conf_level + 1;
prev_err = 0;
if (regFlag)
    conf_val = 6.63;
end
best_clust=[]; best_bic = 0;
while (n < k & sse > err_tar & chi_val > conf_level)
   id = 0;
   max_err = 0;
   sse = 0;
   cur_err = 0;
   first_flag = 1;
   for i=1:n 
      cur_err = clusts{i}.err;
      if (clusts{i}.csz >= nmin2 & cur_err > 0)
         if (regFlag == 0)
            [cl1,cl2, chi_t] = split1(clusts{i});
         else
            [cl1,cl2,chi_t] = reg_split(clusts{i},rbf_prior,full_evid);
         end
         p1 = cl1.csz/clusts{i}.csz;
         p2 = cl2.csz/clusts{i}.csz;
         if (regFlag==0)
            cur_err = cur_err - (cl1.err*p1+cl2.err*p2);
            cur_err = cur_err * (clusts{i}.csz/N);
         else
            cur_err = cur_err - cl1.err - cl2.err;
         end
         if (first_flag == 1)
            max_err = cur_err;
            first_flag = 0;
         end
      end
      flag = 0;
      if ((cur_err >= max_err & clusts{i}.csz >= nmin2))
         max_err = cur_err;
         id = i;
         best_cl1 = cl1;
         best_cl2 = cl2;
         if (conf_level_flag & regFlag==0)
            chi_val = chi_t;
         end
      end
      sse = sse + clusts{i}.sse;
   end
   sse = sse / N;
   if (id == 0)
      break;
   end
   csz = clusts{id}.csz;
   cl1 = best_cl1;
   cl2 = best_cl2;
   clusts{id} = cl1;
   n = n + 1;
   clusts{n} = cl2;
   sse = 0;
   for i=1:n
      sse = sse + clusts{i}.sse;
   end
   if (regFlag==1)
      sse = sse/N;
      net = get_net_from_clusts(net,clusts,p,t,opts);
      y = prbfn_out(p,net);
      prev_err = (y-t)*(y-t)';
      sse = prev_err;
      bic = -N*0.5*log(sse/N) - (n/2)*log(N);
      if (bic > best_bic | n==2)
          best_clust=clusts;
          best_bic = bic;
      end
   end
end
if (regFlag==1 & bic_flag)
    clusts = best_clust;
    n = length(clusts);
end
cnts = [];
rads = [];
cm = [];

for i=1:n
   cm = [cm clusts{i}.csz];
   if (regFlag == 0)
      nc = mean(clusts{i}.p,2);
   else
      nc = clusts{i}.p(:,clusts{i}.max_err_id); 
      nc = clusts{i}.m;
   end
   cnts = [cnts nc];
   if (isfield(clusts{i}, 'rad'))
      r = clusts{i}.rad;
   else
      r = diag(cov(clusts{i}.p));
      r = sqrt(max(r));
   end
   rads = [rads r];
end
cinit = cnts;
if (opts.verbose)
   fprintf(1, 'func_clust:%d clusters sse:%f\n', n, sse);
end

function net = get_net_from_clusts(net,clusts,p,t,opts,cl1,cl2,i)
m = []; r=[];
if (nargin > 5)
    len = length(clusts);
    clusts{i} = cl1;
    clusts{len+1} = cl2;
end
for i=1:length(clusts)
   m = [m clusts{i}.m];
   r = [r clusts{i}.rad];
end
net = create_net_from_clusts(net,clusts,p,t,m,r,opts);

%-------------------------------------------------------
function [cl1, cl2,chi] = split1(clust)
% [cl1, cl2] = split(clust);
%Try to split about the two most prominent classes.
id1=1; id2 = [];
[dimt,N] = size(clust.t);
id1 = find(clust.t(1,:)==1);
maxN = length(id1);
maxN2 = length(id2);
dim1 = 1; dim2 = 0;
if (dimt > 1)
   for d=2:dimt
      nd = find(clust.t(d,:)==1);
      lnd = length(nd);
      if (lnd > maxN)
         id2 = id1;
         dim2 = dim1;
         dim1 = d;
         maxN2 = maxN;
         id1 = nd;
         maxN = lnd;
      elseif (lnd > maxN2)
         id2 = nd;
         maxN2 = lnd;
         dim2 = d;
      end
   end
else
   id1 = find(clust.t==1);
   id2 = find(clust.t==0);
   dim2 = 1;
end
p1 = mean(clust.p(:,id1),2);
p2 = mean(clust.p(:,id2),2);
cl1.p = [];
cl1.t = [];
cl2.p = [];
cl2.t = [];
l1 = length(id1);
l2 = length(id2);
p1 = clust.p(:,id1); p2=clust.p(:,id2);
%[w,wo] = LinCls([p1 p2], [ones(1,l1) (ones(1,l2)*(-1))]);
%[w,wo] = fld([p1 p2], [ones(1,l1) (ones(1,l2)*(-1))]);
%[w,wo] = lin_fld([p1 p2], [ones(1,l1) (ones(1,l2)*(-1))]);
[w,wo] = linfld([p1 p2], [ones(1,l1) (ones(1,l2)*(-1))]);
yvals = w'*clust.p + wo;
id1 = find(yvals > 0);
id2 = find(yvals < 0);
npats = l1 + l2;
cl1.p = clust.p(:,id1);
cl1.t = clust.t(:,id1);
cl2.p = clust.p(:,id2);
cl2.t = clust.t(:,id2);
n1 = length(find(cl1.t(dim1,:)==1));
Prob = n1/npats;
pn1 = Prob * l1;
pn2 = Prob * l2;
if (dim1==dim2)
   n2 = length(find(cl1.t(1,:)==0));
else
   n2 = length(find(cl1.t(dim2,:)==1));
end
if (pn1==0 | pn2 == 0)
   chi = (pn1-n1)^2 +(pn2-n2)^2;
else
   chi = (pn1-n1)^2/pn1 +(pn2-n2)^2/pn2;
end
t1m = mean(cl1.t,2);
[dimt,csz] = size(cl1.t);
cl1.csz = csz;
err = t1m*ones(1,csz) - cl1.t;
err = sum(err.*err,1);
[tmp,cl1.max_err_id] = max(err);
cl1.err = entropy(cl1);
cl1.sse = sum(err);

t2m = mean(cl2.t,2);
[dimt,csz] = size(cl2.t);
cl2.csz = csz;
err = t2m*ones(1,csz) - cl2.t;
err = sum(err.*err,1);
[tmp,cl2.max_err_id] = max(err);
cl2.err = entropy(cl2);
cl2.sse = sum(err);

%-------------------------------------------------------
function h = entropy(clust)
% Computes the entropy 
[dim,csz] = size(clust.t);
prob = zeros(1,dim);
if (dim > 1)
   for i=1:dim
      prob(i) = length(find(clust.t(i,:)==1));
   end
else
   prob(1) = length(find(clust.t==1));
   prob(2) = length(find(clust.t==0));
end
prob = prob/csz;
h = zeros(1,dim);
ids = find(prob>0);
h(ids) = -prob(ids).*log(prob(ids));
h = sum(h);

%-------------------------------------------------------
function [w,wo] = linfld(x,y)
% [w,wo] = linfld(x,y) - computes linear hyper plane
[dim,N] = size(x);
id1 = find(y == -1);
id2 = find(y == 1);
x1 = x(:,id1);
x2 = x(:,id2);
m1 = mean(x1,2);
m2 = mean(x2,2);
n1 = length(id1);
n2 = length(id2);
xx1 = sum(sum(x1.*x1))/n1;
xx2 = sum(sum(x2.*x2))/n2;
m1m2 = (m1+m2)'*(m1+m2)/2;
sxx = xx1 + xx2 - m1m2;
if (sxx > 0)
   w = (m2 - m1)./sxx;
else
   w = (m2 - m1);
   w = w./sqrt(w'*w);
end
wo = -w'*(m1+m2)/2;


%-------------------------------------------------------
function [cl1, cl2,chi_t] = reg_split(clust, rbf_prior,full_evid)
% [cl1, cl2,chi_t] = reg_split(clust);
% Description: This function finds the patterns with the maximum and minmum
% values and split the cluster around them.
p = clust.p;
t = clust.t;
[dim,N] = size(p);
tm = mean(t,2);
[ymax, id_max] = max(t);
[ymin, id_min] = min(t);
p1 = p(:,id_max);
p2 = p(:,id_min);
idx1 = [];
idx2 = [];
for i=1:N
   d1 = p(:,i) - p1;
   d2 = p(:,i) - p2;
   d1 = d1'*d1;
   d2 = d2'*d2;
   if (d1 <= d2)
      idx1 = [idx1 i];
   else
      idx2 = [idx2 i];
   end
end

p1 = p(:,idx1);
t1 = t(:,idx1);
p2 = p(:,idx2);
t2 = t(:,idx2);
cl1.p = p1;
cl1.t = t1;
cl2.p = p2;
cl2.t = t2;

t1m = mean(cl1.t,2);
[dimt,csz] = size(cl1.t);
cl1.csz = csz;
err = t1m*ones(1,csz) - cl1.t;
err = sum(err.*err,1);
%[tmp,cl1.max_err_id] = max(err);
[tmp,cl1.max_err_id] = max(cl1.t);
cl1.sse = sum(err);
[w,m,r] = prbfn_reg_uint(cl1.p,cl1.t,1);
cl1.m = m;
cl1.rad = r;
cl1.w = w;
[rbf_fit,mlp_fit,m,w,r,rbf_err,mlp_err,id1,tmp,flgs] = prbfn_func_fit(cl1.p,cl1.t,1,full_evid);
if (rbf_fit < 0)
   rbf_fit = rbf_fit / rbf_prior;
else
   rbf_fit = rbf_fit * rbf_prior;
end
if (rbf_fit >= mlp_fit)
   cl1.fit ='rbf';
else
   cl1.fit = 'mlp';
end
cl1.flgs = flgs;
cl1.err = cl1.sse;

t2m = mean(cl2.t,2);
[dimt,csz] = size(cl2.t);
cl2.csz = csz;
err = t2m*ones(1,csz) - cl2.t;
err = sum(err.*err,1);
%[tmp,cl2.max_err_id] = max(err);
[tmp,cl2.max_err_id] = min(cl2.t);
cl2.sse = sum(err);
[w,m,r] = prbfn_reg_uint(cl2.p,cl2.t,0);
cl2.m = m;
cl2.rad = r;
cl2.w = w;
[rbf_fit,mlp_fit,m,w,r,rbf_err,mlp_err,id2,tmp,flgs] = prbfn_func_fit(cl2.p,cl2.t,1,full_evid);
if (rbf_fit < 0)
   rbf_fit = rbf_fit / rbf_prior;
else
   rbf_fit = rbf_fit * rbf_prior;
end
if (rbf_fit >= mlp_fit)
   cl2.fit = 'rbf';
else
   cl2.fit = 'mlp';
end
cl2.flgs = flgs;
cl2.err = cl2.sse;
n = clust.csz;
e1 = sqrt(clust.sse/n);
e2 = sqrt((cl1.sse + cl2.sse)/n);
chi_t = 2*n*(log(e1) - log(e2));

%-------------------------------------------------------
function [rbf_fit,mlp_fit,m,w,r,rbf_err,mlp_err,id,nflag,flgs] = prbfn_func_fit(x,y,r,full_evid)
% [rbf_fit,mlp_fit,m,w,r,rbf_err,mlp_err,id,nflag] = prbfn_func_fit(x,y)
% Description: Computes the  function fit to the data for 
% ridge or rbf.
% Author: Shimon Cohen

% Normalize to [0..1]
minY = min(y);
y = y - minY;
maxY = max(y);
if (maxY ~= 0) % & full_evid==0)
   y = y ./ maxY;
end   
if (nargin < 4)
   prn = 0;
end
prn=0;
[rbf1,mlp1,m1,w1,r1,rbf_err1,mlp_err1,id1,flgs1] = dofit1(x,y,r,1,full_evid);
[rbf2,mlp2,m2,w2,r2,rbf_err2,mlp_err2,id2,flgs2] = dofit1(x,y,r,0,full_evid);
%[rbf1,mlp1,m1,w1] = dofit2(x,y,r,1);
%[rbf2,mlp2,m2,w2] = dofit2(x,y,r,0);

if (rbf1 > rbf2)
   rbf_fit = rbf1;
   flgs = flgs1;
   m = m1;
   r = r1;
   rbf_err = rbf_err1;
   id = id1;
   nflag = 1;
   
   mlp_fit = mlp1;
   mlp_err = mlp_err1;
   w = w1;
else
   rbf_fit = rbf2;
   flgs = flgs2;
   m = m2;
   r = r2;
   rbf_err = rbf_err2;
   id = id2;
   nflag = 0;
   
   mlp_fit = mlp2;
   mlp_err = mlp_err2;
   w = w2;  
end
rbf_err = rbf_err*maxY*maxY;
mlp_err = mlp_err*maxY*maxY;
%---------------------------------------------------------------
function [rbf,mlp,m,w,r,rbf_val,mlp_val,id,flgs] = dofit1(x,y,r,flag,full_evid)
if (~flag)   
   y = y * (-1);
   y = y - min(y);
end
[ym, id] = max(y);
[dim,K] = size(x);
[dimr,Kr] = size(r);
m = x(:,id);
r = x - m * ones(1,K);
r = sum(r.*r,2)+eps;
if (K > 1)
   r = r./(K-1);
else
   r = ones(dim,1);
end
r = sqrt(r).*sqrt(dim/2);

%r = compute_rad(x,y,m,r);

mnorm = m'*m;
xbar = mean(x,2); ybar = mean(y);
if (mnorm > 0)
   w = m /sqrt(mnorm);
else
   w = m;
end
wo = ybar - w'*xbar;
w = [wo; w];
rbf = -1; mlp = -1;
flgs = ones(dim,1);
wflgs = flgs;
[rbf_val,rbf, mlp_val,mlp] = compute_sse(x,y,m,r,w, flgs,full_evid);
cur_rbf_val = rbf_val;
cur_mlp_val = mlp_val;



%-------------------------------------------------------------------
function [rbf_val, rbf_evid,mlp_val,mlp_evid] = compute_sse(x,y,m,r,w, flgs,full_evid)
[dim,K] = size(x);
[dimr,Kr] = size(r);
rbf_val = 0;
mlp_val = 0;
yrbf = zeros(1,K);
ymlp = zeros(1,K);
x1 = [ones(1,K); x];
flgs1 = [1; flgs];
for i=1:K
   wx= (w.*flgs1)'*x1(:,i);
   mv = 1/(1+exp(-wx));
   ymlp(i) = mv;
   dv = (x(:,i) - m) ./ r;
   dv = dv.*flgs;
   dv = dv'*dv/2;
   mr = exp(-dv);
   yrbf(i) = mr;
end
if (full_evid)
   [rbf_val,rbf_evid] = compute_fit(x,yrbf,y);
   [mlp_val,mlp_evid] = compute_fit(x,ymlp,y);
else
   rbf_val = sum((yrbf-y).*(yrbf-y));
   mlp_val = sum((ymlp-y).*(ymlp-y));
   rbf_evid = 1/(1+rbf_val);
   mlp_evid = 1/(1+mlp_val);
end
%-------------------------------------------------------------------
function rad = compute_rad(x,y,m,avgRad)
% rad = compute_rad(x,y,id)
% Computes the radius from the data
[dim,N] = size(x);
y = y - min(y);
maxY = max(y);
if (maxY ~= 0)
   y = y./ maxY;
end
maxRad = max(avgRad) * 100;
dist = m * ones(1,N) - x;
dist = dist.*dist;
rads = [avgRad 2*avgRad 4*avgRad];
errvals =[];
for i=1:3
   rc = rads(:,i);
   rc = rc.*rc*2;
   rc = rc * ones(1,N);
   mdist = sum(dist ./ rc); 
   yt = exp(-mdist);
   ydiff = (y-yt).*(y-yt);
   err = sum(ydiff);
   errvals=[errvals err];
end
while (errvals(2)>errvals(1) | errvals(2)>errvals(3))
   if (errvals(1) < errvals(3))
      errvals(2:3) = errvals(1:2);
      rads(:,2:3) = rads(:,1:2);
      rads(:,1) = rads(:,1)*0.5;
      id = 1;
   else
      errvals(1:2) = errvals(2:3);
      rads(:,1:2) = rads(:,2:3);
      id = 3;
      rads(:,3) = rads(:,3) * 2;
   end
   rc = rads(:,id) .* rads(:,id) * 2;
   rc = rc * ones(1,N);
   mdist = sum(dist ./ rc); 
   yt = exp(-mdist);
   ydiff = (y-yt).*(y-yt);
   err = sum(ydiff);
   errvals(id) = err;
   if (mean(rads(:,3)) > maxRad)
      break;
   end
end
xvals=[1 2 4];
[a,b,c,rmin] = parab_fit(xvals,errvals);
rmin = rads(:,1) * rmin;
rads = [rads rmin];
rmin = rmin * ones(1,N);
mdist = sum(dist ./ (rmin.*rmin*2)); 
yt = exp(-mdist);
ydiff = (y-yt).*(y-yt);
err = sum(ydiff);
errvals = [errvals err];
[err,id] = min(errvals);
rad = rads(:,id);

%-------------------------------------------------------------------------------
function [err,evid] = compute_fit(x,y,t)
% [err,evid] = compute_fit(x,y,t)
% Description: Computes the fit of the function in given (xi,yi) and ti
a = 0.1; b = 10;
[dim,n] = size(x);
thr = 1e-5;
iter = 0;
denom = (sum(y.*y)-sum(y)*sum(y)/n);
if (denom ~= 0)
   w = (sum(y.*t) - sum(y)*sum(t)/n)/denom;
   wo = (sum(t) - w*sum(y))/n;
else
   w = 1;
   wo =0;
end
err = (w*y+wo) - t;
err = sum(err.*err);

%a = err/n;
%b = w^2;
prev_err = err+10*thr;
prevLL = 0; LL=10*thr;
%while (abs(LL-prevLL)>thr & iter < 15)
while (iter < 15)
   syy = sum(y.*y); sy = sum(y); syt=sum(y.*t);
   denom = (b*syy-(b/n)*sy*sy+a);
   if (denom==0)
      denom = eps;
   end
   if (n>1 & denom ~= 0)
      nom = b*syt - b*sum(t)*sy/n;
      wt = nom/denom;
      if (wt==0)
         break;
      else
         w = wt;
      end
      wo = (sum(t) - w*sum(y))/n;
   elseif (n==1)
      w = 1;
      wo = 0;
      break;
   else
      err = -1;
      return;
   end
   prev_err = err;
   err = (w*y+wo) - t;
   err = sum(err.*err);
   prev_a = a; prev_b = b;
   a = err/n; 
   b = w*w;
   if (b==0 | a==0)
%      disp('b==0 | a==0');
      break;
   end	   
   prevLL = LL;
   LL = -n*log(a) - log(b);
   if (iter == 0)
      prevLL = LL - 10*thr;
   end
   iter = iter + 1;
end
small_num = eps*eps*eps*eps;
if (b==0)
   b = small_num;
end
if (a==0)
   a = small_num;
end
asy = sum(y)/a;
H = [sum(y.*y)/a+1/b asy; asy n/a];
detH = det(H);
if (detH==0)
   detH = small_num;
end
logdetH = log(detH);
if (n>1)
   evid = -0.5*n*log(a) - 0.5*log(b) - 0.5*logdetH;
   %evid = -0.5*n*log(a) - 0.5*log(b);
else
   evid = 0;
end
%err = err + w*w ; evid = 1.0/(err+1);
%---------------------------------------------------------------
function [rbf,mlp,m,w] = dofit2(x,y,r,flag)
if (flag)
   [maxy,id] = max(y);
   y = (y - maxy)*(-1);
else
   y = y - min(y);
   [maxy,id] = min(y);
end
[dim,N] = size(x);
[dimr,Kr] = size(r);
m = x(:,id);
if (dimr == 1 & N > 1)
   xd = m*ones(1,N) - x;
   xd = sum(xd.*xd);
   xdist = xd;
   xd = xd.*xd;
   a = sum(y.*xd)/sum(xd.*xd);
   yrbf = a*xd;
   w = m / sqrt(m'*m);
   xd = w'*x;
   a = (N*sum(xd.*y) -sum(xd)*sum(y))/(N*sum(xd.*xd)-sum(xd)*sum(xd));
   b = (sum(y) - a*sum(xd)) / N;
   ymlp = a*xd + b;
   rbf_val = sum((yrbf-y).*(yrbf-y));
   mlp_val = sum((ymlp-y).*(ymlp-y));
else
   rbf = 1;
   mlp = 0;
end
rbf_val = compute_fit(x,yrbf,y);
mlp_val = compute_fit(x,ymlp,y);
rbf = 1/(1+rbf_val);
mlp = 1/(1+mlp_val);

function [clust_mem, clust_mem_idx] = get_clust_mem(p,m,rads)
% clust_mem = get_clust_mem(p,m)
% Description: returns a structure of matrixes for each center
% Each matrix includes the members of the cluster.
%
% Copyrights: Shimon Cohen
[dim,N] = size(p);
if (nargin < 3)
   rad_flag = 0;
else
   rad_flag = 1;
   rads = rads .* rads;
end
[dim, n] = size(m);
pclust = zeros(1, N);
one_vec = ones(1, n);
clust_mem{n} = [];
clust_mem_idx{n} =[ ];
for ii=1:N
   dist = p(:,ii) * one_vec;
   dist = dist - m;
   dist = dist .* dist;
   if (rad_flag)
      dist = dist ./ rads;
   end
   dist = sum(dist , 1);
   [min_dist, idx] = min(dist);
   clust_mem{idx} = [clust_mem{idx}, p(:,ii)];
   clust_mem_idx{idx} = [clust_mem_idx{idx} ii];
end

function [rr_net,err] = prbfn_create_net(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts,flgs)
% prbfn_create_net - creates the actual net, the sturct given to it
% should contain: output_type, hid_type,err_type,vec_rad.
% These fields contains:
% output_type the type of output: 0-logsig,1-linear
% hid_type: type of hidden unit: 0-logsig, 1-tanh
% err_type: 1 absolute, 2 sse
% vec_rad - 0 for 1 dim raduises, otherwise diagonal matrix.
%
% Author: Shimon Cohen Copyrights(C)

[dimr, nr] = size(r);
rr_net.vec_rad = (dimr > 1);
rr_net.wgts_drv = 0;
M2 = length(rbf_inds);
rr_net.m = m(:, rbf_inds);
rr_net.rads = r(:,rbf_inds);
if (rr_net.vec_rad)
   rr_net.rads = 1 ./ rr_net.rads;
end
rr_net.nridges = length(mlp_inds);
rr_net.nrbfs = M2;

rr_net.rbf_ones = ones(1, M2);

% prepare rmin
rmin = min(rr_net.rads) / 10;
rr_net.rmin = rmin;
[dim,npats] = size(p);
rr_net.p1 = [ones(1, npats); p];
[rr_net,err] = get_init_wgts(rr_net,p,t,m,r, mlp_inds,rbf_inds,opts);
M1 = rr_net.nridges;
if (rr_net.vec_rad)
   rr_net.rmin = 0;
end
if (nargin == 9)
   if (rr_net.nrbfs > 0)
      rr_net.rads = rr_net.rads .* flgs(:,rbf_inds);
   end
   [nr,nc] = size(rr_net.w1);
   rr_net.w1_flgs = ones(nr,nc);
   rr_net.w1_flgs(:,2:end) = flgs(:,mlp_inds)';
end

function [rr_net,cerr] = get_init_wgts(net,p,t,m,r,mlp_idxs,rbf_idxs,opts)
% rr_net = get_init_wgts(net,p,t,m,r,mlp_idxs,rbf_idxs,opts)
% get_init_wgts - gets the initial weights for the
% net. It uses the following criterion: E / norm(gradE)
% The geuss which has lowest criterion value is the best.
% If the gradient is large then there is a lot to enhance, so
% the criterion value will be low. This is the same if
% the error is small in the first place. If the norma
% of the gradient is small then there is not a lot for
% improvement and the error gets larger.
% Inputs:
% net - the rr net
% ntrys - the number of tries to use
% p - the pattern matrix
% t - the target vector
% Outputs:
% rr_net - the ridge rbf net 
%rr_net = prbfn_init_wgts(net,p,t,m,r,mlp_idxs,opts);
%return;
[dimr, nr] = size(r);
net.rmin = eps;
M1 = net.nridges;
M2 = net.nrbfs;
[dim, tmp] = size(p);
err = 10;
[C, npats] = size(t);
class_flag = (C > 1);
if (C == 1)
   id1 = find(t == 1);
   id0 = find(t == 0);
   if ((length(id0) + length(id1) ) == npats)
      class_flag = 1;
   end
end
%class_flag = 0;
%r= r.*r;
net.w1 = zeros(M1, dim+1);
if (class_flag)
	net.w1 = calculate_w1(p,t,m,r, mlp_idxs);   
else	
   % Regression problems
   %   [clust_mems, idxs] = get_clust_mem_rads(p,m,r);
   for i=1:M1         
      id = mlp_idxs(i);
      nv = m(:,id);
      if (dim > 1)
         mv = nv'*nv;         
         mv = sqrt(mv);
         nv = nv ./ mv;         
      end
      net.w1(i,2:dim+1) = nv;
      net.w1(i,1) = 0;
   end
   M2 = length(rbf_idxs);
end
%net.w1 = ann_init_wgts(dim,M1);
net.w2 = ones(C, M1 + M2 + 1);
rr_net = net;
%rr_net = compute_rads(rr_net,p,t);
[y,z,a1,a2] = prbfn_out(p, rr_net,0);
cerr = sumsqrerr(t, y, rr_net.err_type);

if (isfield(opts,'reg2') & opts.reg2>0)
   H = z*z';
   [dimH,dimH2] = size(H);
   H = H + eye(dimH) * opts.reg2;
   pinvZ = z'*pinv(H);
   rr_net.w2 = t*pinvZ';
else
   rr_net.w2 = t*pinv(z)';
end
[y,z,a1,a2] = prbfn_out(p, rr_net,0);
cerr = sumsqrerr(t, y, rr_net.err_type);

% check if the next if is good for regression problems
if (dimr > 1 | class_flag==0)
%   rr_net.w2(:,2:end) = rr_net.w2(:,2:end) * 0.99;
end
if (M2 == 0)
   rr_net.m = [];
   rr_net.rads = [];
end

%------------------------------------------------------------------------------------
function w1 = calculate_w1(p,t, m,r, mlp_idxs)
% w1 = calculate_w1(p,t,m)
% Description: calculate the first weights matrix values
[clust_mems, idxs] = get_clust_mem_rads(p,m,r);
%m = m(:,mlp_idxs);
[dim, K] = size(m);
[dim, N] = size(p);
K = length(mlp_idxs);
for ii=1:K
	real_idxs{ii} = idxs{mlp_idxs(ii)};
end

w1 = zeros(K, dim+1);
[tdim,npats] = size(t);
% Go through each center and calculate their sum
for ii=1:K
   pc = p(:, real_idxs{ii});
   tc = t(:, real_idxs{ii});
   [dim_pc, npc] = size(p);
   if (tdim > 1)
	   for d=1:tdim
   		nc(d) = length(find(tc(d,:) == 1));   
      end
      [maxv,best_d] = max(nc);
  		tv = ones(dim, npats) .* (-1);
	   id = find(t(best_d,:) == 1);
   else
      nc(1) = length(find(tc == 0));
      nc(2) = length(find(tc == 1));
      if (nc(1) > nc(2))
         maxv = 0;
      else
         maxv = 1;
      end
  		tv = ones(dim, npats) .* (-1);
	   id = find(t == maxv);
   end
   %   id = idxs{ii};
   id = real_idxs{ii}; 
   ni = length(id);
   if (ni > 0)
      ni_fac = (ni-N)/ni;
   else
      ni_fac = -1;
   end
   tv = ones(dim+1, npats) .* (-1);
   pt = [ones(1,npats); p];
   tv(:,id) = tv(:,id) .* (ni_fac);
   pt = pt .* tv;
   
   v = mean(pt,2);
   %------------------
  % v = [1; m(:,ii)];
  % v = v ./ (v'*v);
   %----------------
	v(1) = 0;
 %  maxv = max(abs(v));
% 	v = v.*3/maxv;
   v = v./sqrt(v'*v);
   w1(ii, :) = v';
end	

%---------------------------------------------------------
function [clust_mems, idxs] = get_clust_mem_rads(p,m,r)
% [clust_mems, idxs] = get_clust_mem(p,m,r)
% Description: associate evry pattern with the right cluster
[dimr,K] = size(r);
[dim, K] = size(m);
[dim, N] = size(p);
clust_mems{K} = [];
idxs{K} = [];
if (dimr == 1)
   r = ones(dim,K);
end
ones_mat = ones(1,K);
r2 = r .* r;
for i = 1:N
   diff = m - p(:,i) * ones_mat;
   diff = (diff .* diff) ./ r2;
   dist = sum(diff);
   [min_dist, id] = min(dist);
   clust_mems{id} = [clust_mems{id} p(:,i)];
   idxs{id} = [idxs{id} i];
end


function [a,b,c, xmin] = parab_fit(x,y)
% [a,b,c] = parab_fit(x, y)
% description: fits a parabola ax^2 + bx  + c and returns the parameters
% Author: Shimon Cohen
% Copyrights(C) Shimon Cohen
min_dist = 1e-15;
if (x(2)-x(1) > min_dist & (x(3)-x(2) > min_dist))
   a=(y(2)-y(1))/(x(2)-x(1)) - (y(3)-y(1))/(x(3)-x(1));
   a = a / (x(2)-x(3));
   b = (y(2)-y(1))/(x(2)-x(1)) - a * (x(1) + x(2));
   c = y(1) - a * x(1) * x(1) - b * x(1);
else
   a = 0;
   b = 0;
   c = 0;
end
if (a > 0)
   xmin = - b/(2*a);
else
   xmin = -1;
end

%-----------------------------------------------------
function [rr_net,iter,rc,err] = prbfn_conj_grad(net, p, t, max_iter, err_tar, opts)
% rr_net = prbfn_conj_grad(net, p, t)
% Description: given a net and p,t train according to the
% conjugate train algorithm.
% Inputs:
% net - the input params of the net.
% p - the pattern matrix
% t - the target values
% opts - regularization options
% Outputs:
% rr_net - the trained net
rr_net = net;
rmin = rr_net.rmin;
rinit = net.rads;
cancel_dr = 0;
verbose = 0;
if (isfield(opts, 'verbose'))
   verbose = opts.verbose;
end
if (isfield(opts, 'cancel_dr'))
   cancel_dr = opts.cancel_dr;
end
% number of of weights, or the size of gradient
if (rr_net.wgts_drv == 0)
   grad = [rr_net.w1(:); rr_net.w2(:); rr_net.m(:); rr_net.rads(:)];
else
   grad = [rr_net.w1(:); rr_net.w2(:)];
end
[wgts_cnt,tmp] = size(grad);

% start beta value for the first conjugate direction
beta = 0;
err_eps = 0.00000001;
min_lr = 0.0000001;
min_lr = eps;

dedm = 0; dedr = 0;
disp_freq = 50;
if (isfield(opts, 'disp_freq'))
   disp_freq = opts.disp_freq;
end
if (isfield(opts, 'min_rad_flag'))
   min_rad_flag = 1;
else
   min_rad_flag = 0;
end
min_rad_flag=1;
%if (max_iter < 10)
%   max_iter = 10 * wgts_cnt;
%end
lr = 0.1;
%lr = 1;
max_lr = lr * 2;
orig_lr = lr;
if (rr_net.err_type == 1)
   orig_lr = 0.001;
end
prev_lr = orig_lr;
iter = 0;

[y,z,a1,a2] = prbfn_out(p, rr_net,0);
err = sumsqrerr(t, y, rr_net.err_type);
prev_err = err;
min_flag = 1;
if (verbose)
   fprintf(1, 'Conjugate gradient prbfn_train max_iter:%d\n', max_iter);
   fprintf(1, 'Start error=%f, net.output_type=%d,hid_type=%d\n', err, rr_net.output_type,rr_net.hid_type);
end
nmis = 1;
rc = 1;
M1 = rr_net.nridges;
M2 = rr_net.nrbfs;
while (iter < max_iter & err >= err_tar & nmis > 0 & rc==1)
   [dedw1, dedw2, dedm, dedr] = prbfn_drv(p, t, rr_net, a1, a2, z, y);
   %      [dedw11, dedw21, dedm1, dedr1] = prbfn_finite_diff(rr_net,p, t);
   if (cancel_dr)
      dedr = dedr .* 0;
   end
   if (opts.reg1 ~= 0)
      dedw1 = dedw1 + opts.reg1 * rr_net.w1;
   end
   if (opts.reg2 ~= 0)
      dedw2(:,2:M1+1) = dedw2(:,2:M1+1) + opts.reg2 * rr_net.w2(:,2:M1+1);
   end
   if (opts.reg3 ~= 0)
      dedw2(:,M1+2:end) = dedw2(:,M1+2:end) + opts.reg3 * rr_net.w2(:,M1+2:end); 
   end
   if (opts.regr ~= 0 & rr_net.nrbfs > 0)
      if (rr_net.vec_rad==0)
         dr =  2*opts.regr .* rr_net.rads;
         dedr = dedr - dr;
      else
         dedr = dedr + opts.regr*rr_net.rads;
      end
   end
   new_grad = [dedw1(:); dedw2(:); dedm(:); dedr(:)];
   
   % computes the direction
   gg = grad'*grad;
   if (abs(gg) < eps)
       if (verbose)
         fprintf(1,'grad norm too small local min gg=%f\n', gg);
       end
       break;
   end
   if (gg == 0)
       fprintf(1,'Bugggggggggg gg = 0\n');
   end 
   if (mod(iter, wgts_cnt) ~= 0)
      beta = new_grad'*(new_grad-grad)/gg;
      beta = max(0,beta);
      dedw1 = beta * old_dedw1 - dedw1;
      dedw2 = beta * old_dedw2 - dedw2;
      dedm  = beta * old_dedm - dedm;
      dedr  = beta * old_dedr - dedr;
   else
      % zeros previuos direction
      dedw1 = -dedw1;
      dedw2 = -dedw2;
      dedm  = -dedm;
      dedr  = -dedr;         
   end
   
   % line search for best learn rate at the 
   % given conjugate direction
   lr = orig_lr;
   
   rr_net.rmin = net.rmin;
   [lr,min_flag,err] = line_search(rr_net,p, t, dedw1, dedw2, dedm,dedr, lr, min_lr, max_lr,prev_err,err_tar);
 %  lr = fmin('prbfn_update_net',0,3,[0,1.e-16], p,t,rr_net, dedw1, dedw2,dedm,dedr);
   if (lr == 0)
      rc = 0;
   end
   % [lr,min_flag,err] = rbf_mlp_line_search(rr_net,p, t, dedw1, dedw2, dedm,dedr, lr, min_lr, max_lr,prev_err,err_tar);
   %[err,lr]=prbfn_line_search(rr_net,p,t,dedw1,dedw2,dedm,dedr,min_lr,max_lr);
   rr_net.w1 = rr_net.w1 + lr * dedw1;
   rr_net.w2 = rr_net.w2 + lr * dedw2;
   rr_net.rads = abs(rr_net.rads + lr * dedr);
   
   if (rr_net.rmin > 0)
      bad_r = find(rr_net.rads < rmin);
      if (~isempty(bad_r))
         rr_net.rads = abs(rr_net.rads);
      end	
      
      bad_r = find(rr_net.rads < rmin);
      if (min_rad_flag == 1)
         rr_net.rads(bad_r) = rmin;
         bad_r = [];
      end
      if (~isempty(bad_r))
         fprintf(1, 'rad go to zero****\n');
         rr_net.rads = rr_net.rads - lr * dedr;
         rr_net.w1 = rr_net.w1 - lr * dedw1;
         rr_net.w2 = rr_net.w2 - lr * dedw2;
         rr_net.wgts_drv = 1;
         dedm=0; dedr=0;
         new_grad = [dedw1(:); dedw2(:)];
      else
         rr_net.m = rr_net.m + lr * dedm;   
      end
   else
      rr_net.m = rr_net.m + lr * dedm; 
   end
   
   [y,z,a1,a2] = prbfn_out(p, rr_net,0);
   err = sumsqrerr(t,y, rr_net.err_type);
      
   % update values for next iteration, gradient
   % and search direction
   grad = new_grad;
   old_dedw1 = dedw1;
   old_dedw2 = dedw2;
   old_dedm  = dedm;
   old_dedr  = dedr;
   
   iter = iter + 1;
   
   if (mod(iter, disp_freq) == 0 & verbose)
      fprintf(1, 'iter=%d, err=%f, gg=%f, lr=%f, beta=%f, nmis=%d\n', iter, err,gg,lr, beta, nmis);      
      
      %      nmis = get_num_mis(y,t);
      %      if (nmis == 0)
      %         break;
      %      end
   end
   prev_err = err;
end

if (verbose == 1)
   [dimr,rl] = size(rr_net.rads);
   if (rl > 0)
      avg_r = sum(rr_net.rads)/rl;
      if (dimr > 1)
         avg_r = sum(avg_r)/dimr;
      end
   else
      avg_r = 0;
   end
   fprintf(1, 'conj_grad_train end error:%f, iter:%d,lr:%f,nmis:%d,avg_r:%f\n', err,iter,lr,nmis,avg_r);
   if (err > err_tar)
      fprintf(1, 'Error goal hasnt been reached, increase or decrease \n number of hidden units\n');
   end
%   prbfn_disp_stat(p,t,z, rr_net);
end

function [lr,min_flag,err] = line_search(net, p, t, dedw1, dedw2, dedm, dedr, lrc, min_lr, max_lr, prev_err, err_tar)
% lr = line_search(net, dedw1, dedw2, dedm, dedr,err,lr, lr_min, lr_max)
% performs a line search in order to find the best lr.
% It uses parabola approximation in order to find the minmum at each stage
% Thus saving computation time for the forward computation
% of he netwrok outputs.
max_iter = 100;
iter = 1;
lr_dec = 0.5;
lr_inc = 2;
lr = lrc * lr_dec;
rr_net = net;
min_flag = 0;
lrvals(1)=0;
lrvals(3) = max_lr;
if (rr_net.err_type == 1)
   lrvals(3) = 0.0001;
end
lrvals(2) = (lrvals(3) + lrvals(1)) * 0.5;
errvals(1) = prev_err;
lrvals(1) = lrvals(2) * 0.5;
for i=1:3
   lr = lrvals(i);
   [net, errvals(i)] = update_net(p,t,rr_net, lr, dedw1,dedw2,dedm,dedr);
end
% search on left side
while (errvals(2) >= errvals(1) | errvals(2) >= errvals(3) | isnan(errvals(3)) | errvals(2) > prev_err)
   [a,b,c] = parab_fit(lrvals, errvals);
   if (a > 0)
      lrc = - b/(2*a);
   else
      lrc = -1;
   end
   if (lrc > 0)
      [net, err] = update_net(p,t,rr_net, lrc, dedw1,dedw2,dedm,dedr);
      if (err < errvals(2))
         if (lrc > lrvals(2))
            if (lrc < lrvals(3))
               id = 2;
            else
               id = 3;
            end
         else
            if (lrc < lrvals(1))
               id = 1;
            else
               id = 2;
            end
         end
         lrvals(id) = lrc;
         errvals(id) = err;
      end
      if (lrc < lrvals(2) & err < errvals(1))
         lrvals(1) = lrc;
         errvals(1) = err;
      end
      if (lrc > lrvals(2) & err < errvals(3))
         lrvals(3) = lrc;
         errvals(3) = err;
      end
      if (errvals(2) < errvals(1) & errvals(2) < errvals(3) & errvals(2) < prev_err)
         break;
      end
   end
   %  | errvals(3) > prev_err was removed from the if 
   if (errvals(3) >= errvals(1) | isnan(errvals(3)) | errvals(3) > prev_err)
      lrvals(2:3) = lrvals(1:2);
      errvals(2:3) = errvals(1:2);
      lrvals(1) = lrvals(1) * 0.5;
      lr = lrvals(1);
      [net, errvals(1)] = update_net(p,t,rr_net, lr, dedw1,dedw2,dedm,dedr);
   else 
      % search on the right side
      lrvals(1:2) = lrvals(2:3);
      errvals(1:2) = errvals(2:3);
      lrvals(3) = lrvals(3) * lr_inc;
      lr = lrvals(3);
      [net, errvals(3)] = update_net(p,t,rr_net, lr, dedw1,dedw2,dedm,dedr);
   end
   iter = iter + 1;
   if (iter > max_iter)
      [err,idx] = min(errvals);
      lr = lrvals(idx);
      return;
   end
end
[a,b,c] = parab_fit(lrvals, errvals);
cnt = 0;
if (a > 0)
   while (cnt <= 3 & min_flag == 0 & a > 0)
      cnt = cnt + 1;
      lrc = - b/(2*a);
      if (lrc > 0)
         [net, err] = update_net(p,t,rr_net, lrc, dedw1,dedw2,dedm,dedr);
         if (err <= errvals(2))
            lr = lrc;
            lrvals(2) = lr;
            errvals(2) = err;
            min_flag = 1;
         else
            if (lrc > lrvals(2))
               errvals(3) = err;
               lrvals(3) = lrc;
            else
               errvals(1) = err;
               lrvals(1) = lrc;
            end
            [a,b,c] = parab_fit(lrvals, errvals);
         end
      else
         break;
      end
   end
end   
if (min_flag == 0)
   [err,ind] = min(errvals);
   lr = lrvals(ind);
end

%------------------------------------------------------
function [net, err] = update_net(p,t,rr_net, lr, dedw1, dedw2,dedm,dedr)
[err,net] = prbfn_update_net(lr,p,t,rr_net, dedw1, dedw2,dedm,dedr);

%--------------------------------------------------------------------------
function [net,iter]=prbfn_vconj_grad(rr_net, p, t, max_iters, err_tar, opts)
% [rr_net,iter] = prbfn_vconj_grad(rr_net,p,t,max_iter,err_tar,opts)
% Description: Using a validation set train the classifier to get the best
% performance on the validation set.
% The valildation set is usually 1/3 of the set. After training and getting the best 
% error on the validation, the validation is added to the train set and a re-train
% is done.
% Author: Shimon Cohen
% Copyrights (C)
maxtry = 10;
verr = -1;
[dim,N] = size(p);
perm = randperm(N);
train_size = ceil(N*2/3);
pt = p(:,perm(1:train_size));
tt = t(:,perm(1:train_size));
pv = p(:,perm(train_size+1:N));
tv = t(:,perm(train_size+1:N));
[dim,npats] = size(pt);
rr_net.p1 = [ones(1, npats); pt];
max_iters1 = 10;
if (isfield(opts, 'max_iters1'))
    max_iters1 = opts.max_iters1;
end
iters = 0;
if (max_iters < 0)
    max_iters = length(rr_net.w1(:)) + length(rr_net.m(:));
end
while (iters < max_iters)
   [rr_net,iter,rc,cur_train_err] = prbfn_conj_grad(rr_net, pt, tt, max_iters1, 0, opts);
   iters = iters + max_iters1;
	[y,z,a1,a2] = prbfn_out(pv, rr_net,1);
	err = sumsqrerr(tv, y, rr_net.err_type);
   if (err < verr | verr < 0)
      net = rr_net;
      verr = err;
      train_err = cur_train_err;
   end
end
net.p1 = [ones(1, N); p];
[net,iter,rc,err] = prbfn_conj_grad(net, p, t, max_iters, train_err, opts);


%-------------------------------------------------------
function [err, nsse, terr] = sumsqrerr(t,y, err_type)
% [err, nsse, terr] = sumsqrerr(t, y, err_type)
% description: Returns the sum sequare error between 
% targets and outputs. The error returned is normalized
% by the number of patterns and output classes.
%
% Inputs:
% t - targets in column matrix
% y - outputs in colomnu matrix
% Output:
% err - error value normalized by the number of outputs.
% nsse - the error normalized by the data targets variance
% terr - the total error
%
% Author: Shimon Cohen
[ncls, npats] = size(t);
delta = t - y;
if (err_type == 1)
   delta = abs(delta);
elseif (err_type == 2)
   delta = delta .* delta;
else
   sig = 10;
   delta = tanh(delta/sig);
   delta = delta.*delta;
end
err = sum(sum(delta));

if (ncls > 1)
   tmean = mean(t,2) * ones(1,npats);
else
   tmean = mean(t);
end
var = (t - tmean) .* (t - tmean);
var = sum(sum(var));
if (var==0)
   disp('sumsqrerr var is zero');
   var=1;
end
nsse = err / var;
terr = err;
err = err / npats;

function [dedw1,dedw2,dedm,dedr] = prbfn_drv(p, t, rr_net, a1, a2, z, y)
% [dedw1,dedw,dedm,dedr]=prbfn_drv(p,t, rr_net, a1, a2, z, y)
% Description: This function computes the partial derviatives
% of the network in respect to w1, w2 , m and r. 
% Inputs:
% p - a pattern matrix
% t - a target matrix
% rr_net - the net params
% a1 - product of wieghts(w1) and input vector
% a2 - product of weights and outputs of first layer
% z - outputs of first layer
% y - output of outer layer (actually output of the net)
% Outputs:
% dedw1 - the partial derivitives in respect to w1
% dedw2 - partial derivitivs in respect to w2
% dedm - partial derevitivs in respect to centers
% dedr - partial derivitivs in respect to radiuses
% Author: Shimon Cohen
[dim, npats] = size(p);
[rdim, nrs] = size(rr_net.rads);

if (rr_net.err_type == 1)
   delta = sign(y - t);
elseif (rr_net.err_type==2)
   delta = y - t;
else
   sig = 10;
   del = tanh(y - t);
   delta = 1 - del.*del;
end

M1 = rr_net.nridges;
M2 = rr_net.nrbfs;

[C, nm] = size(rr_net.w2);
M = rr_net.nrbfs + rr_net.nridges;

% derivation of g with respect to ak
ones_Cn = ones(C, npats);
if (rr_net.output_type == 2)
   gak_prim = ones_Cn;
elseif (rr_net.output_type == 1)
   gak_prim = y .* (ones_Cn - y);
elseif (rr_net.output_type == 3)   
   gak_prim = 1 - y .* y;
end

% computes delta of outer layer (CxN)
del_kn = gak_prim .* delta;

% compute partial drv in respect to outer weights
dedw2 = del_kn * z;

% compute derivatives in respect to w1 z = (Nx(C+1))
zj = z(:,2:M1+1);
if (rr_net.hid_type == 1)
   gaj_prim = zj .* (1 - zj);
else
   gaj_prim = 1 - zj .* zj;
end

%dedw1 = zeros(M1, dim+1);

% w2=(Cx(M+1)), del_kn=(CxN), w1=(NRidge x (dim+1))
dedw1_tmp = del_kn' * rr_net.w2(:,2:M1+1);
% dedw1_tmp = (NxNRidges), gaj_prim=(N x NRidges)
dedw1_tmp = gaj_prim .* dedw1_tmp;
% dedw1_tmp = (NxNRidges) p = (dim x N)
dedw1 = dedw1_tmp' * rr_net.p1';

% Computes derivatives in repsect to centers and radiues
phee = z(:,2+rr_net.nridges:end);

% phee = (N x NRbfs)
dedm = zeros(dim, M2);
dedr = zeros(rdim, M2);
if (M2 <= 0)
   dedm=rr_net.m; 
   dedr=rr_net.rads;
   return;
end
% del_kn=(CxN) w2=(Cx(M+1)) dedmj=(dim x M2)
% del_kn_mul_w2 = (N x M2)
del_kn_mul_w2 = del_kn' * rr_net.w2(:,M1+2:end);

% Now each radial has a number to multiply with
ones_dim1 = ones(dim,1);
if (rr_net.vec_rad == 0)
   rmat = ones_dim1 * rr_net.rads;
   rmat2 = rmat .* rmat;
else
   rmat2 = rr_net.rads .* rr_net.rads;;
end

rads3 = rr_net.rads .* rr_net.rads;
rads3 = rr_net.rads .* rads3;
ones_1M2 = ones(1,M2);
for i=1:npats
   x = p(:,i) * ones_1M2;
   diff_mat = x - rr_net.m;
   
   if (rr_net.vec_rad == 0)
      diff_mat2 = diff_mat ./ rmat2;
      diff_mat3 = sum(diff_mat .* diff_mat) ./ rads3;
   else
      diff_mat2 = diff_mat .* rmat2;
      diff_mat3 = (diff_mat .* diff_mat) .* rr_net.rads;
      diff_mat3 = -diff_mat3;
   end
      
   % phee = (N x M2), del_kn_mul_w2=(NxM2),wgtj = (1 x M2)
   wgtj = del_kn_mul_w2(i,:) .* phee(i,:);
   if (rr_net.vec_rad == 1)
      wgtj = ones_dim1 * wgtj;
   end
   dedrj = wgtj .* diff_mat3;
   dedr = dedr + dedrj;
   
   if (rr_net.vec_rad == 0)
      % diff_mat2=(dim x M2), wgtj = (1 x M2)
      wgtj = ones_dim1 * wgtj;
   end
   % wgtj=(dim x M2) diff_mat2=(dim x M2)
   dedmj = wgtj .* diff_mat2;
   dedm = dedm + dedmj;
end
% Remove pruned derivatives
if (isfield(rr_net, 'w2_flgs'))
   dedw2 = dedw2 .* rr_net.w2_flgs;
   if (~isempty(rr_net.w1))
      dedw1 = dedw1 .* rr_net.w1_flgs;
   end
end
if (isempty(rr_net.m))
   dedm = rr_net.m;
   dedr = rr_net.rads;
end
if (isempty(rr_net.w1))
   dedw1 = rr_net.w1;
end

function [err,net] = prbfn_update_net(lr,p,t,rr_net, dedw1, dedw2,dedm,dedr)
% [net,err] = update_net(rr_net,lr,dedw1,dedw2,dedm, dedr)
% Description: updating the net accordig to derivatives
% and learn rate.
net = rr_net;
net.w1 = rr_net.w1 + lr * dedw1;
net.w2 = rr_net.w2 + lr * dedw2;
rmin = rr_net.rmin;

if (rr_net.wgts_drv == 0 & rr_net.nrbfs > 0)
   net.m = rr_net.m + lr * dedm;
   net.rads = rr_net.rads + lr * dedr;
   net.rads = abs(net.rads);
   if (rr_net.rmin > 0)
      bad_r = find(abs(net.rads) < rmin);
      net.rads(bad_r) = rmin;
   end
else
   net.m = rr_net.m;
   net.rads = rr_net.rads;
end   
[y,z,a1,a2] = prbfn_out(p, net,0);
err = sumsqrerr(t,y, net.err_type);

function [w,m,r] = prbfn_reg_uint(x,y,max_flg)
% [w,m,r] = prbfn_reg_uint(x,y,m, max_flg)
%  Description:returns the initial values for a given cluster. The max_flg
% denotes if woprking on maximum of minmum.
%
% Author: Shimon Cohen
minY = min(y);
y = y - minY;
maxY = max(y);
if (maxY ~= 0) 
   y = y ./ maxY;
end   
if (~max_flg)   
   y = y * (-1);
   y = y - min(y);
end 
[ym, id] = max(y);
[dim,K] = size(x);
m = x(:,id);
r = x - m * ones(1,K);
r = sum(r.*r,2)+eps; 
if (K > 1)
   r = r./(K-1);
else
   r = ones(dim,1);
end 
r = sqrt(r).*sqrt(dim/2);
mnorm = m'*m;
xbar = mean(x,2); ybar = mean(y); 
if (mnorm > 0)
   w = m /sqrt(mnorm);
else
   w = m;
end 
wo = ybar - w'*xbar;
w = [wo; w]; 

function [onet, ensemble] = prbfn_prune(p,t, inet, opts)
% [onet,ensamble] = prbfn_prune(p,t, inet, max_iter, err_tar, opts)
% Description: Prunes one hidden unit at each and continue this process until
% the prune until error target cannot be reached.
% opts.model_sel_type: BIC = 2, SSE/(n-p) =1, LRT(0.95), LRT(0.975),LRT(0.99)
% 
% Copyrights(C) Shimon Cohen
[nr,nc] = size(inet.w2);
inet.w2_flgs = ones(nr,nc);
max_nhids = inet.nrbfs;
nparams = nr*nc;
[rdim,nK] = size(inet.rads);
% BIC is the default
model_sel = 2;
if (isfield(opts,'model_sel_type'))
   model_sel = opts.model_sel_type;
end
if (nK>1 & rdim == 1)
   [dim,nK] = size(inet.m);
   inet.rads = ones(dim,1)*inet.rads;
   inet.vec_rad = 1;
end
if (inet.nridges > 0)
   [nr,nc] = size(inet.w1);
   inet.w1_flgs = ones(nr,nc);
   nparams = nparams + nr*nc;
end
if (inet.nrbfs > 0)
   [nr,nc] = size(inet.m);
   nparams = nparams + nr*nc;
end
nparams = get_param_num(inet);
ens_size = 0;
if (isfield(opts,'ensemble_size'))
   ens_size = opts.ensemble_size;
end
ensemble.size = 0;
ensemble.errors = [];
ensmeble.nets=[];
nhids = inet.nrbfs + inet.nridges;
[dimt,N] = size(t);
max_iter = 0;
if (isfield(opts, 'max_iters'))
   max_iter = opts.max_iters;
end
if (opts.prune_val > 0)
   err_tar = 0;
else
   err_tar = opts.err_tar;
end
[inet,iter,rc,err] = prbfn_conj_grad(inet,p,t,max_iter, err_tar, opts);
onet = inet;
cur_err = err;
err_tar = opts.err_tar;
base_err = opts.err_tar;
err_tar = cur_err;
max_iter = opts.max_iters1;
rc = 1;
nprn = 0;
max_params = nparams;
if (opts.prune_val > 1)
   min_params = opts.prune_val;
   max_prn = max_params - min_params;
else
   max_prn = round(nparams*opts.prune_val);
   min_params = max_params - max_prn;
end
prev_err = err_tar;
prev_params = nparams;
thr = 3.841;
[err_tar,cont_flg] = compute_model_err(err_tar, nparams, N, model_sel,p,t,inet,prev_err,prev_params,thr,opts.err_tar);
cur_err = err_tar;
best_err = cur_err;
verbose = 0;
if (isfield(opts,'verbose'))
   verbose = opts.verbose;
end
while (rc>0 & min_params < nparams & nprn < max_prn & cont_flg)
   err_tar = cur_err;
   if (ens_size > 0)
      ensemble = add_net(ensemble, inet, cur_err,ens_size);
   end
   if (cur_err < best_err | (cont_flg==1 & model_sel>=4))
      onet = inet;
      best_err = cur_err;
   end
   [inet,rc] = prune(inet,p,t,opts);
   if (rc < 0)
      break;
   end
   prev_err = err;
   [inet,iter,rc,err] = prbfn_conj_grad(inet,p,t,max_iter, prev_err, opts);
%   nparams = nparams - 1;
   prev_params = nparams;
   nparams = get_param_num(inet);
   [cur_err,cont_flg,reg] = compute_model_err(err, nparams,N, model_sel,p,t,inet,prev_err,prev_params,thr,opts.err_tar);
   if (reg > 0)
      inet.reg2 = reg;
      inet.reg3 = reg;
   end
   nprn = max_params - nparams;
   if (verbose)
      fprintf(1,'pruned = %d, nparams = %d\n',nprn, nparams);
   end
end
nprms = get_param_num(onet);
if (verbose)
   fprintf(1,'max_params=%d, nparams(after pruning)=%d,err_tar=%f,cur_err=%f\n', max_params, nprms,err_tar,cur_err);
end

%--------------------------------------------------------------
function np = get_param_num(net)
np = 0;
if (net.nridges>0 & isfield(net, 'w1_flgs'))
   np = np + length(find(net.w1_flgs>0));
end
if (net.nrbfs > 0 & isfield(net, 'rads'));
   np = np + 2*length(find(net.rads>0));
end
np = np + length(find(net.w2_flgs>0));

%--------------------------------------------------------------
function [err,cont_flg,reg] = compute_model_err(err , nparams, N, model_sel,p,t,net,prev_err,prev_params,thr,error_tar)
cont_flg = 1;
reg = -1;
if (model_sel == 1)
   % Computes MSE = SSE/(N-p);
   np = N - nparams;
   if (np < 0)
      np = eps;
   end
   err = N*err /np;
elseif (model_sel == 2)
   % Computes the BIC
   %err = N*err * 0.5 + 0.5 * nparams * log(N); 
   %   err = N * log(sqrt(err)) + 0.5 * nparams * log(N);
   err = 0.5* N * log(err) + 0.5 * nparams * log(N);
elseif (model_sel == 3)
   [evid,reg] = prbfn_glm_evidence(p,t,net);
   err = -evid;
elseif (model_sel == 4 | model_sel == 5 | model_sel == 6)
   if (model_sel == 4)
      thr1 = 3.84;
      thr2 = 5.99;
   elseif (model_sel == 5)
      thr1 = 5.02;
      thr2 = 7.38;
   else
      thr1 = 6.63;
      thr2 = 9.21;
   end
   df = prev_params-nparams;
   conf = 2*N*(log(sqrt(err)) - log(sqrt(prev_err)));
   cont_flg = 0;
   if (conf <= thr1 & (df==1 | df ==0))
      cont_flg = 1;
   elseif (conf <= thr2 & df==2)
      cont_flg = 1;
   end
else
   % Use error tar to stop prune   
   if (err > error_tar)
      cont_flg = 0;
   end
end

%------------------------------------------------------------
function ensemble = add_net(ensemble, inet, cur_err,ens_size)
if (ensemble.size < ens_size)
   ensemble.errors = [ensemble.errors cur_err];
   ensemble.size = ensemble.size + 1;
   ensemble.nets{ensemble.size} = inet;
else
   maxi = 0; max_err=0;
   for i=1:ensemble.size
      if (maxi==0)
         maxi = i;
         max_err = ensemble.errors(i);
      elseif (ensemble.errors(i) > max_err)
            max_err = ensemble.errors(i);
            maxi = i;
      end
   end
   if (max_err > cur_err)
      ensemble.nets{maxi} = inet;
      ensemble.errors(maxi) = cur_err;
   end
end

%--------------------------------------------
function dv = getdet(H)
i = 1;
dv = det(H);
[d,d]=size(H);
im = eye(d);
while (dv == 0)
   i = i * 10;
   dv = det(H + im *i*eps);
end

%--------------------------------------------------------------
function [net,rc] = prune(net,p,t, opts)
r1 = -1;
c1 = -1;
errs = [-1 -1 -1];
if (net.nridges > 0 & sum(sum(net.w1_flgs))>0)
   [r1,c1,errs(1)] = prune_w1(net,p,t,opts);
end
if (sum(sum(net.rads)))
   [r2,c2,errs(2)] = prune_w2(net,p,t,opts);
end
if (net.nrbfs > 0 & any(net.rads))
   [r3,c3,errs(3)] = prune_rads(net,p,t,opts);
end
mid = 0;
min_err = -1;
for i=1:3
   if (errs(i) >0)
      if (errs(i) < min_err | min_err <0)
         mid = i;
         min_err = errs(i);
      end
   end
end
rc = -1;
if (mid > 0)
   rc = 1;
   if (mid == 1)
      net.w1_flgs(r1,c1) = 0;
   elseif (mid == 2)
      net.w2_flgs(r2,c2) = 0;
   elseif (mid == 3)
      net.rads(r3,c3) = 0;
      if (sum(net.rads(:,c3)) == 0)
         net.rads(:,c3) = [];
         net.m(:,c3) = [];
         net.w2(:,net.nridges+1+c3) = [];
         net.w2_flgs(:,net.nridges+1+c3) = [];
         net.nrbfs = net.nrbfs - 1;
      end
   end
end
%-------------------------------------------------
function [r,c,min_err] = prune_w1(net,p,t,opts)
[nr,nc] = size(net.w1);
min_err = -1;
r=-1;c=-1;
for ir=1:nr
   for ic=1:nc
      if (net.w1_flgs(ir,ic)>0)
         net.w1_flgs(ir,ic) = 0;
         y = prbfn_out(p, net,0);
         err = compute_err(y,t,net.err_type);
         if (err < min_err | min_err < 0)
            min_err = err;
            r = ir;
            c = ic;
         end
         net.w1_flgs(ir,ic) = 1;
      end
   end
end

%-------------------------------------------------
function [r,c,min_err] = prune_w2(net,p,t,opts)
[nr,nc] = size(net.w2);
min_err = -1;
r=-1;c=-1;
for ir=1:nr
   for ic=1:nc
      if (net.w2_flgs(ir,ic)>0)
         net.w2_flgs(ir,ic) = 0;
         y = prbfn_out(p, net,0);
         err = compute_err(y,t,net.err_type);
         if (err < min_err | min_err < 0)
            min_err = err;
            r = ir;
            c = ic;
         end
         net.w2_flgs(ir,ic) = 1;
      end
   end
end

%-------------------------------------------------
function [r,c,min_err] = prune_rads(net,p,t,opts)
[nr,nc] = size(net.rads);
min_err = -1;
for ir=1:nr
   for ic=1:nc
      if (net.rads(ir,ic) > 0)
         rv = net.rads(ir,ic);
         net.rads(ir,ic) = 0;
         y = prbfn_out(p, net,0);
         err = compute_err(y,t,net.err_type);
         if (err < min_err | min_err < 0)
            min_err = err;
            r = ir;
            c = ic;
         end
         net.rads(ir,ic) = rv;
      end
   end
end

function err = compute_err(y,t, err_type)
[dimt,N] = size(y);
err = y - t;
err = sum(sum(err.*err))/N;